// Copyright (c) 2013 by István Váradi

// This file is part of VSCPL, a simple cross-platform utility library

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

#ifndef HU_VARADIISTVAN_SCPL_IO_WIN32_LOCALACCEPTOR_H
#define HU_VARADIISTVAN_SCPL_IO_WIN32_LOCALACCEPTOR_H
//------------------------------------------------------------------------------

#include "../BufferedStreamAcceptor.h"

#include "Completer.h"

#include "LocalServerSocketBase.h"
#include "LocalSocket.h"

//------------------------------------------------------------------------------

namespace hu { namespace varadiistvan { namespace scpl { namespace io {

//------------------------------------------------------------------------------

/**
 * Instances of this class are used to accept connection requests to a
 * local server socket, i.e. named pipe.
 */
class LocalAcceptor : public Completer,
                      public BufferedStreamAcceptor
{
private:
    /**
     * The path of the pipe to accept.
     */
    char path[256];

    /**
     * Indicate if the connection has been accepted.
     */
    bool accepted;

protected:
    /**
     * Construct the acceptor.
     */
    LocalAcceptor(LocalServerSocketBase* socket, const char* name);

    /**
     * Get the socket.
     *
     * It returns the overlappable from the Completer part, knowing
     * that it must be a LocalServerSocketBase.
     */
    LocalServerSocketBase& getServerSocket() const;

public:
    /**
     * Try to accept a connection.
     *
     * If one is already accepted, but not retrieved yet, return true.
     *
     * If a new connection becomes immediately available, return true.
     *
     * If a new connection is not yet available, returns false.
     *
     * If some error occurs, the acceptor is marked failed and false
     * is returned.
     */
    bool accept();

    /**
     * Get the local socket accepted last. If no socket was accepted,
     * return 0. The local socket's waiter will be the given one.
     */
    LocalSocket* getSocket(Waiter* waiter,
                           size_t readingCapacity = LocalSocket::DEFAULT_CAPACITY,
                           size_t writingCapacity = LocalSocket::DEFAULT_CAPACITY);
    /**
     * Get the local socket accepted last. If no socket was accepted,
     * return 0. The local socket's waiter will be the same as that of
     * the associated server socket.
     */
    LocalSocket* getSocket(size_t readingCapacity = LocalSocket::DEFAULT_CAPACITY,
                           size_t writingCapacity = LocalSocket::DEFAULT_CAPACITY);


    /**
     * Get the last accepted socket as a buffered stream with the given waiter,
     * if any.
     */
    BufferedStream* getAcceptedBufferedStream(Waiter* waiter = nullptr) override;

protected:
    /**
     * Set the indicator of an accepted connection to true.
     *
     * @see Completer::handleWaitedResult
     */
    virtual void handleWaitedResult(DWORD size);

    friend class LocalServerSocket;
};

//------------------------------------------------------------------------------
// Inline definitions
//------------------------------------------------------------------------------

inline LocalServerSocketBase& LocalAcceptor::getServerSocket() const
{
    return static_cast<LocalServerSocketBase&>(overlappable);
}

//------------------------------------------------------------------------------

inline LocalSocket* LocalAcceptor::getSocket(Waiter* waiter,
                                             size_t readingCapacity,
                                             size_t writingCapacity)
{
    if (!accepted) return 0;

    LocalSocket* s = new LocalSocket(waiter, overlappable.releaseHandle(),
                                     readingCapacity, writingCapacity);
    accepted = false;
    return s;
}

//------------------------------------------------------------------------------

inline LocalSocket* LocalAcceptor::getSocket(size_t readingCapacity,
                                             size_t writingCapacity)
{
    return getSocket(overlappable.getWaiter(),
                     readingCapacity, writingCapacity);
}

//------------------------------------------------------------------------------

} /* namespace hu::varadiistvan::scpl::io */ } /* namespace hu::varadiistvan::scpl */ } /* namespace hu::varadiistvan */ } /* namespace hu */

//------------------------------------------------------------------------------
#endif // HU_VARADIISTVAN_SCPL_IO_WIN32_LOCALACCEPTOR_H

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
